Attribute VB_Name = "MHtmlClipRoutines"
' **************************************************************************
'  Copyright 2002 Karl E. Peterson
'  All Rights Reserved, http://www.mvps.org/vb
' **************************************************************************
'  You are free to use this code within your own applications, but you
'  are expressly forbidden from selling or otherwise distributing this
'  source code, non-compiled, without prior written consent.
' **************************************************************************
'  Additional Sources:
'  http://support.microsoft.com/default.aspx?scid=KB;en-us;q274326
'  http://msdn.microsoft.com/workshop/networking/clipboard/htmlclipboard.asp
' **************************************************************************
Option Explicit

' Win32 API declarations
Private Declare Function RegisterClipboardFormat Lib "user32" Alias "RegisterClipboardFormatA" (ByVal lpString As String) As Long
Private Declare Function OpenClipboard Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function GetClipboardData Lib "user32" (ByVal wFormat As Long) As Long
Private Declare Function SetClipboardData Lib "user32" (ByVal wFormat As Long, ByVal hMem As Long) As Long
Private Declare Function GetPriorityClipboardFormat Lib "user32" (lpPriorityList As Long, ByVal nCount As Long) As Long
Private Declare Function CloseClipboard Lib "user32" () As Long
Private Declare Function GlobalAlloc Lib "kernel32" (ByVal wFlags As Long, ByVal dwBytes As Long) As Long
Private Declare Function GlobalLock Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Function GlobalUnlock Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (pTo As Any, uFrom As Any, ByVal lSize As Long)
Private Declare Function lstrlenA Lib "kernel32" (ByVal lpString As Long) As Long
Private Declare Function lstrlenW Lib "kernel32" (ByVal lpString As Long) As Long

' Predefined Clipboard Formats
Private Const CF_TEXT = 1
Private Const CF_BITMAP = 2
Private Const CF_METAFILEPICT = 3
Private Const CF_SYLK = 4
Private Const CF_DIF = 5
Private Const CF_TIFF = 6
Private Const CF_OEMTEXT = 7
Private Const CF_DIB = 8
Private Const CF_PALETTE = 9
Private Const CF_PENDATA = 10
Private Const CF_RIFF = 11
Private Const CF_WAVE = 12
Private Const CF_UNICODETEXT = 13
Private Const CF_ENHMETAFILE = 14
Private Const CF_HDROP = 15
Private Const CF_LOCALE = 16
Private Const CF_MAX = 17

Private Const CF_OWNERDISPLAY = &H80
Private Const CF_DSPTEXT = &H81
Private Const CF_DSPBITMAP = &H82
Private Const CF_DSPMETAFILEPICT = &H83
Private Const CF_DSPENHMETAFILE = &H8E

' Used to obtain registered format for HTML
Private CF_HTML As Long
Private Const RegHtml As String = "HTML Format"

Public Sub Main()
   ' This sub intended for routine testing.
   ' Comment out to include in other projects.
   
   ' Get system value for HTML format.
   CF_HTML = RegisterClipboardFormat(RegHtml)
   Debug.Print Hex$(CF_HTML)
   
   ' Dump HTML contents to Immediate Window.
   Debug.Print HtmlDescription(GetText(CF_HTML))
End Sub

Public Function Html2Text(ByVal Data As String) As String
   Dim obj As Object
   ' Hijack an IE offering...
   Screen.MousePointer = vbHourglass
   On Error Resume Next
   Set obj = CreateObject("htmlfile")
   With obj
      .Open
      .Write Data
      Html2Text = .Body.InnerText
   End With
   Screen.MousePointer = vbDefault
End Function

Public Function HtmlAttribute(ByVal Data As String, ByVal Attr As String) As String
   Dim Begins As Long
   Dim CrLf As Long
   
   ' Search for first occurance of Attribute name.
   On Error GoTo Bail
   Begins = InStr(1, Data, Attr, vbTextCompare)
   If Begins Then
      ' Should be immediately followed by a colon.
      If Mid$(Data, Begins + Len(Attr), 1) = ":" Then
         Begins = Begins + Len(Attr) + 1
         ' Find terminating Cr/Lf.
         CrLf = InStr(Begins, Data, vbCrLf)
         ' Extract attribute value.
         HtmlAttribute = Mid$(Data, Begins, CrLf - Begins)
      End If
   End If
Bail:
End Function

Public Function HtmlContext(ByVal Data As String) As String
   Dim Starts As Long
   Dim Ends As Long
   
   ' If numbers are off, errors may occur.
   On Error Resume Next
   
   ' Find starting/ending positions of HTML.
   ' Offsets are zero-based, so we add one to satisfy VB.
   Starts = CLng(Val(HtmlAttribute(Data, "StartHTML"))) + 1
   Ends = CLng(Val(HtmlAttribute(Data, "EndHTML"))) + 1
   
   ' Return the HTML, minus its description.
   HtmlContext = Mid$(Data, Starts, Ends - Starts)
End Function

Public Function HtmlDescribed(ByVal Fragment As String) As String
   Dim Data As String
   Dim nPos As Long
   Const Description As String = _
      "Version:1.0" & vbCrLf & _
      "StartHTML:aaaaaaaaaa" & vbCrLf & _
      "EndHTML:bbbbbbbbbb" & vbCrLf & _
      "StartFragment:cccccccccc" & vbCrLf & _
      "EndFragment:dddddddddd" & vbCrLf
   Const FragmentStart As String = "<!--StartFragment-->"
   Const FragmentEnd As String = "<!--EndFragment-->"
   Const Fmt As String = "0000000000"
   
   ' Add the starting and ending tags for the HTML fragment
   ' by looking for <body> tag.
   nPos = InStr(1, Fragment, "<body", vbTextCompare)
   Select Case nPos
      Case 0
         Fragment = "<html><body>" & vbCrLf & FragmentStart & Fragment
      Case Else
         nPos = InStr(nPos, Fragment, ">")
         If nPos > 0 And nPos < Len(Fragment) Then
            Fragment = Left$(Fragment, nPos) & FragmentStart & Mid$(Fragment, nPos + 1)
         End If
   End Select
      
   nPos = InStr(1, Fragment, "</body", vbTextCompare)
   Select Case nPos
      Case 0
         Fragment = Fragment & FragmentEnd & vbCrLf & "</body></html>"
      Case Else
         Fragment = Left$(Fragment, nPos - 1) & FragmentEnd & Mid$(Fragment, nPos)
   End Select
   
   ' Build the HTML given the description, the fragment and the context.
   ' And, replace the offset place holders in the description with values
   ' for the offsets of StartHMTL, EndHTML, StartFragment and EndFragment.
   ' Offsets need to be zero-based when placed on clipboard, so subtract 1
   ' from each before injecting.
   Data = Description & Fragment
   Data = Replace(Data, "aaaaaaaaaa", Format$(Len(Description), Fmt))
   Data = Replace(Data, "bbbbbbbbbb", Format$(Len(Data), Fmt))
   nPos = InStr(Data, FragmentStart) - 1
   Data = Replace(Data, "cccccccccc", Format$(nPos + Len(FragmentStart), Fmt))
   nPos = InStr(Data, FragmentEnd) - 1
   Data = Replace(Data, "dddddddddd", Format$(nPos, Fmt))

   ' Return attributed string.
   HtmlDescribed = Data
End Function

Public Function HtmlDescription(ByVal Data As String) As String
   Dim Starts As Long
   
   ' If numbers are off, errors may occur.
   On Error Resume Next
   
   ' Find starting/ending positions of HTML.
   Starts = CLng(Val(HtmlAttribute(Data, "StartHTML")))
   
   ' Return the HTML, minus its description.
   HtmlDescription = Left$(Data, Starts)
End Function

Public Function HtmlFragment(ByVal Data As String) As String
   Dim Starts As Long
   Dim Ends As Long
   
   ' If numbers are off, errors may occur.
   On Error Resume Next
   
   ' Find starting/ending positions of HTML.
   ' Offsets are zero-based, so we add one to satisfy VB.
   Starts = CLng(Val(HtmlAttribute(Data, "StartFragment"))) + 1
   Ends = CLng(Val(HtmlAttribute(Data, "EndFragment"))) + 1
   
   ' Return the HTML, minus its description.
   HtmlFragment = Mid$(Data, Starts, Ends - Starts)
End Function

' **********************************************************
'  Utility Routines
'  From ClipView.zip on  http://www.mvps.org/vb/samples.htm
' **********************************************************
Public Function GetText(Optional ByVal Format As Long) As String
   Dim nFmt As Long
   Dim hData As Long
   Dim lpData As Long
   
   If Format = 0 Then
      ' Nothing specific asked, retrieve best text available.
      nFmt = GetPriorityFormat(CF_TEXT, CF_UNICODETEXT, CF_OEMTEXT, CF_DSPTEXT)
   Else
      nFmt = Format
   End If
   
   ' -1=None requested, 0=Empty
   If nFmt > 0 Then
      ' Grab text from clipboard, if available.
      If OpenClipboard(0&) Then
         hData = GetClipboardData(nFmt)
         ' Slurp characters from global memory.
         If hData Then
            lpData = GlobalLock(hData)
            If nFmt = CF_UNICODETEXT Then
               GetText = PointerToStringW(lpData)
            Else
               GetText = PointerToStringA(lpData)
            End If
            Call GlobalUnlock(hData)
         End If
         Call CloseClipboard
      End If
   End If
End Function

Public Function SetText(ByVal NewVal As String, Optional ByVal Format As Long) As Boolean
   Dim hData As Long
   Dim lpData As Long
   Dim Buffer() As Byte
   Const GMEM_FIXED As Long = &H0
   
   ' Try to set text onto clipboard.
   If OpenClipboard(0&) Then
      ' Convert data to ANSI byte array.
      Buffer = StrConv(NewVal & vbNullChar, vbFromUnicode)
      ' Allocate enough memory for buffer.
      hData = GlobalAlloc(GMEM_FIXED, UBound(Buffer) + 1)
      If hData Then
         ' Copy data to alloc'd memory.
         lpData = GlobalLock(hData)
         Call CopyMemory(ByVal lpData, Buffer(0), UBound(Buffer) + 1)
         Call GlobalUnlock(hData)
         ' Hand data off to clipboard
         SetText = CBool(SetClipboardData(Format, hData))
      End If
      Call CloseClipboard
   End If
End Function

Private Function GetPriorityFormat(ParamArray Formats()) As Long
   Dim Fmts() As Long
   Dim i As Long
   Dim nFmt As Long
   
   ' Bail, if no formats were requested.
   If UBound(Formats) < 0 Then Exit Function
   
   ' Transfer desired formats into a non-variant array.
   ReDim Fmts(0 To UBound(Formats)) As Long
   For i = 0 To UBound(Formats)
      ' Double conversion, to be safer.
      ' Could error trap, but that'd mean the
      ' user was a hoser, and we wouldn't want
      ' to insinuate *that*, would we? <g>
      Fmts(i) = CLng(Val(Formats(i)))
   Next i
   
   ' Try opening clipboard...
   If OpenClipboard(0&) Then
      ' Check to see which format is highest in list.
      nFmt = GetPriorityClipboardFormat(Fmts(0), UBound(Fmts) + 1)
      Call CloseClipboard
   Else
      ' Clipboard may already be open by another
      ' routine in same process, try anyway to see
      ' if we can get a successful result.  Not
      ' clean, but worth a shot.
      nFmt = GetPriorityClipboardFormat(Fmts(0), UBound(Fmts) + 1)
   End If
   
   ' Return results
   GetPriorityFormat = nFmt
End Function

Private Function PointerToStringA(ByVal lpStringA As Long) As String
   Dim Buffer() As Byte
   Dim nLen As Long
   
   If lpStringA Then
      nLen = lstrlenA(ByVal lpStringA)
      If nLen Then
         ReDim Buffer(0 To (nLen - 1)) As Byte
         CopyMemory Buffer(0), ByVal lpStringA, nLen
         PointerToStringA = StrConv(Buffer, vbUnicode)
      End If
   End If
End Function

Private Function PointerToStringW(ByVal lpStringW As Long) As String
   Dim Buffer() As Byte
   Dim nLen As Long

   If lpStringW Then
      nLen = lstrlenW(lpStringW) * 2
      If nLen Then
         ReDim Buffer(0 To (nLen - 1)) As Byte
         CopyMemory Buffer(0), ByVal lpStringW, nLen
         PointerToStringW = Buffer
      End If
   End If
End Function

